# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import sys
import textwrap

import click

from eodag.api.core import EODataAccessGateway
from eodag.utils.exceptions import UnsupportedProvider
from eodag.utils.logging import setup_logging


# disable warning on Python 2
click.disable_unicode_literals_warning = True

# A list of supported crunchers that the user can choose (see --cruncher option below)
CRUNCHERS = [
    'RemoveDoubles', 'FilterLatestByName', 'FilterLatestIntersect', 'FilterOverlap',
]


@click.group()
@click.option('-v', '--verbose', count=True,
              help='Control the verbosity of the logs. For maximum verbosity, type -vvv')
@click.pass_context
def eodag(ctx, verbose):
    if ctx.obj is None:
        ctx.obj = {}
    ctx.obj['verbosity'] = verbose


@eodag.command(name='search',
               help='Search satellite images by their product types, and optionally crunch the search results before '
                    'storing them in a geojson file')
@click.option('-b', '--geometry', type=(float,) * 4, default=(None,) * 4,
              help='Search for a product on a bounding box, providing its minlon, minlat, maxlon and maxlat (in this '
                   'order)')
@click.option('-s', '--startTimeFromAscendingNode',
              help='Maximum age of the product (in ISO8601 format: yyyy-MM-ddThh:mm:ss.SSSZ)')
@click.option('-e', '--completionTimeFromAscendingNode',
              help='Minimum age of the product (in ISO8601 format: yyyy-MM-ddThh:mm:ss.SSSZ)')
@click.option('-c', '--cloudCover', type=click.IntRange(0, 100),
              help='Maximum cloud cover percentage needed for the product')
@click.option('-f', '--conf', help='File path to the user configuration file with its credentials',
              type=click.Path(exists=True))
@click.option('-p', '--productType',
              help='The product type to search')
@click.option('--cruncher', type=click.Choice(CRUNCHERS), multiple=True,
              help='A cruncher to be applied to search results. Repeat option many times to apply many crunchers')
@click.option('--cruncher-args', type=(str,) * 3, multiple=True,
              help='Named arguments acting as the parameters of a cruncher. '
                   'Enter it like this: --cruncher-args <CruncherName> <arg-name> <arg-value>. Repeat option many '
                   'times to give many args to a cruncher')
@click.option('--storage', type=click.Path(dir_okay=False, writable=True, readable=False),
              default='search_results.geojson',
              help='Path to the file where to store search results (.geojson extension will be automatically appended '
                   'to the filename). DEFAULT: search_results.geojson')
@click.pass_context
def search_crunch(ctx, **kwargs):
    # Process inputs for search
    kwargs['verbose'] = ctx.obj['verbosity']
    setup_logging(**kwargs)
    if kwargs['geometry'] != (None,) * 4:
        rect = kwargs.pop('geometry')
        footprint = {'lonmin': rect[0], 'latmin': rect[1], 'lonmax': rect[2], 'latmax': rect[3]}
    else:
        footprint = None
    criteria = {
        'geometry': footprint,
        'startTimeFromAscendingNode': kwargs.pop('starttimefromascendingnode'),
        'completionTimeFromAscendingNode': kwargs.pop('completiontimefromascendingnode'),
        'cloudCover': kwargs.pop('cloudcover'),
    }
    producttype = kwargs.pop('producttype')
    conf_file = kwargs.pop('conf')
    if conf_file:
        conf_file = click.format_filename(conf_file)
    if not producttype:
        with click.Context(search_crunch) as ctx:
            print('Give me some work to do. See below for how to do that:', end='\n\n')
            click.echo(search_crunch.get_help(ctx))
        sys.exit(-1)

    # Process inputs for crunch
    cruncher_names = set(kwargs.pop('cruncher') or [])
    cruncher_args = kwargs.pop('cruncher_args')
    cruncher_args_dict = {}
    if cruncher_args:
        for cruncher, argname, argval in cruncher_args:
            cruncher_args_dict.setdefault(cruncher, {}).setdefault(argname, argval)

    satim_api = EODataAccessGateway(user_conf_file_path=conf_file)

    # Search
    results = satim_api.search(producttype, **criteria)
    click.echo("Found {} products with product type '{}': {}".format(len(results), producttype, results))

    # Crunch !
    for cruncher in (satim_api.get_cruncher(cname, **cruncher_args_dict.get(cname, {})) for cname in cruncher_names):
        results = results.crunch(cruncher, **criteria)

    storage_filepath = kwargs.pop('storage')
    if not storage_filepath.endswith('.geojson'):
        storage_filepath += '.geojson'
    result_storage = satim_api.serialize(results, filename=storage_filepath)
    click.echo("Results stored at '{}'".format(result_storage))


@eodag.command(name='list', help='List supported product types')
@click.option('-p', '--provider', help='List product types supported by this provider')
@click.pass_context
def list_pt(ctx, **kwargs):
    kwargs['verbose'] = ctx.obj['verbosity']
    setup_logging(**kwargs)
    dag = EODataAccessGateway()
    provider = kwargs.pop('provider')
    text_wrapper = textwrap.TextWrapper()
    click.echo('Listing available product types:')
    try:
        for product_type in dag.list_product_types(provider=provider):
            text_wrapper.initial_indent = '\n* {}: '.format(product_type['ID'])
            text_wrapper.subsequent_indent = ' ' * len(text_wrapper.initial_indent)
            click.echo(text_wrapper.fill(product_type['desc'] or 'No description'))
    except UnsupportedProvider:
        click.echo('Unsupported provider. You may have a typo')
        click.echo('Available providers: {}'.format(', '.join(dag.available_providers())))
        sys.exit(1)


@eodag.command(help='Download a list of products from a serialized search result')
@click.option('--search-results', type=click.Path(exists=True, dir_okay=False),
              help='Path to a serialized search result')
@click.option('-f', '--conf', type=click.Path(exists=True),
              help='File path to the user configuration file with its credentials', )
@click.pass_context
def download(ctx, **kwargs):
    search_result_path = kwargs.pop('search_results')
    if not search_result_path:
        with click.Context(download) as ctx:
            click.echo('Nothing to do (no search results file provided)')
            click.echo(download.get_help(ctx))
        sys.exit(0)
    kwargs['verbose'] = ctx.obj['verbosity']
    setup_logging(**kwargs)
    conf_file = kwargs.pop('conf')
    if conf_file:
        conf_file = click.format_filename(conf_file)
        satim_api = EODataAccessGateway(user_conf_file_path=conf_file)
        search_results = satim_api.deserialize(search_result_path)
        for downloaded_file in satim_api.download_all(search_results):
            if downloaded_file is None:
                click.echo('A file may have been downloaded but we cannot locate it')
            else:
                click.echo('Downloaded {}'.format(downloaded_file))


@eodag.command(help='Start eodag rpc server')
@click.option('-h', '--host', type=click.STRING, default='localhost',
              help='Interface where to listen for requests')
@click.option('-p', '--port', type=click.INT, default=50051,
              help='The port where to listen for requests', )
@click.option('-f', '--conf', type=click.Path(exists=True),
              help='File path to the user configuration file with its credentials', )
@click.pass_context
def serve(ctx, host, port, conf):
    setup_logging(verbose=ctx.obj['verbosity'])
    from eodag.rpc.server import EODAGRPCServer
    server = EODAGRPCServer(host, port, conf)
    server.serve()


if __name__ == '__main__':
    eodag(obj={})



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import requests
from requests import HTTPError

from eodag.plugins.authentication.base import Authentication
from eodag.utils import RequestsDictTokenAuth, RequestsTextTokenAuth


class TokenAuth(Authentication):

    def authenticate(self):
        # First get the token
        response = requests.post(
            self.config['auth_uri'],
            data=self.config['credentials']
        )
        try:
            response.raise_for_status()
        except HTTPError as e:
            raise e
        else:
            if self.config.get('token_type', 'text') == 'json':
                return RequestsDictTokenAuth(response.json(), self.config['token_key'])
            return RequestsTextTokenAuth(response.text)



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import logging
from operator import attrgetter

import geojson
from pkg_resources import resource_filename
from tqdm import tqdm

from eodag.api.search_result import SearchResult
from eodag.config import SimpleYamlProxyConfig
from eodag.plugins.instances_manager import PluginInstancesManager
from eodag.utils import ProgressCallback
from eodag.utils.exceptions import PluginImplementationError, UnsupportedProvider


logger = logging.getLogger('eodag.core')


class EODataAccessGateway(object):
    """An API for downloading a wide variety of geospatial products originating from different types of systems.

    :param user_conf_file_path: Path to the user configuration file
    :type user_conf_file_path: str or unicode
    :param providers_file_path: Path to the internal file where systems containing eo products are configured
    :type providers_file_path: str or unicode
    """

    def __init__(self, user_conf_file_path=None, providers_file_path=None):
        self.providers_config = SimpleYamlProxyConfig(resource_filename('eodag', 'resources/providers.yml'))
        self.product_types_config = SimpleYamlProxyConfig(resource_filename('eodag', 'resources/product_types.yml'))
        if providers_file_path is not None:
            # TODO : the update method is very rudimentary by now => this doesn't work if we are trying to override a
            # TODO (continues) : param within an instance configuration
            self.providers_config.update(SimpleYamlProxyConfig(providers_file_path))
        if user_conf_file_path:
            self.user_config = SimpleYamlProxyConfig(user_conf_file_path)

            # Override system default config with user values for some keys
            for instance_name, instance_config in self.user_config.items():
                if isinstance(instance_config, dict):
                    if instance_name in self.providers_config:
                        if 'credentials' in instance_config:
                            if 'api' in self.providers_config[instance_name]:
                                self.providers_config[instance_name]['api'].update(instance_config)
                            else:
                                auth_conf = self.providers_config[instance_name].setdefault('auth', {})
                                auth_conf.update(instance_config)
                        for key in ('outputs_prefix', 'extract'):
                            if key in self.user_config:
                                user_spec = self.user_config[key]
                                if 'api' in self.providers_config[instance_name]:
                                    default_dl_option = self.providers_config[instance_name]['api'].setdefault(
                                        key,
                                        '/data/satellites_images/' if key == 'outputs_prefix' else True)
                                else:
                                    if 'download' in self.providers_config[instance_name]:
                                        default_dl_option = self.providers_config[instance_name].get(
                                            'download', {}
                                        ).setdefault(
                                            key,
                                            '/data/satellites_images/' if key == 'outputs_prefix' else True
                                        )
                                    else:
                                        default_dl_option = user_spec  # allows skipping next if block
                                if default_dl_option != user_spec:
                                    if 'api' in self.providers_config[instance_name]:
                                        self.providers_config[instance_name]['api'][key] = user_spec
                                    else:
                                        self.providers_config[instance_name]['download'][key] = user_spec
        self._plugins_manager = PluginInstancesManager(self.providers_config)
        self._plugins_cache = {}

    def set_preferred_provider(self, provider):
        """Set max priority for the given provider.

        >>> import eodag.utils.exceptions
        >>> dag = EODataAccessGateway()
        >>> dag.get_preferred_provider()    # This also tests get_preferred_provider method by the way
        ('peps', 1)
        >>> # For the following lines, see http://python3porting.com/problems.html#handling-expected-exceptions
        >>> try:
        ...     dag.set_preferred_provider(u'unknown')
        ...     raise AssertionError(u'UnsupportedProvider exception was not raised as expected')
        ... except eodag.utils.exceptions.UnsupportedProvider:
        ...     pass
        >>> dag.set_preferred_provider(u'USGS')
        >>> dag.get_preferred_provider()
        ('USGS', 2)
        >>> dag.set_preferred_provider(u'theia')
        >>> dag.get_preferred_provider()
        ('theia', 3)
        >>> dag.set_preferred_provider(u'USGS')
        >>> dag.get_preferred_provider()
        ('USGS', 4)

        :param provider: The name of the provider that should be considered as the preferred provider to be used for
                         this instance
        :type provider: str or unicode
        """
        if provider not in self.available_providers():
            raise UnsupportedProvider('This provider is not recognised by eodag')
        preferred_provider, max_priority = self.get_preferred_provider()
        if preferred_provider != provider:
            new_priority = max_priority + 1
            self.providers_config[provider]['priority'] = new_priority
            # Update the interfaces cache to take into account the fact that the preferred provider has changed
            if 'search' in self._plugins_cache:
                for search_interfaces in self._plugins_cache['search'].values():
                    for iface in search_interfaces:
                        if iface.instance_name == provider:
                            iface.priority = new_priority

    def get_preferred_provider(self):
        """Get the provider currently set as the preferred one for searching products, along with its priority.

        :return: The provider with the maximum priority and its priority
        :rtype: tuple(str, int)
        """
        # Note: if a provider config doesn't have 'priority' key, it is considered to have minimum priority (0)
        preferred, config = max(((provider, conf) for provider, conf in self.providers_config.items()),
                                key=lambda item: item[1].get('priority', 0))
        return preferred, config.get('priority', 0)

    def list_product_types(self, provider=None):
        """Lists supported product types.

        :param provider: The name of a provider that must support the product types we are about to list
        :type provider: str or unicode
        :returns: The list of the product types that can be accessed using eodag.
        :rtype: list(dict)
        :raises: :class:`~eodag.utils.exceptions.UnsupportedProvider`
        """
        if provider is not None:
            if provider in self.providers_config:
                provider_supported_products = self.providers_config[provider]['products']
                return [dict(
                    ID=code,
                    **self.product_types_config[code]
                ) for code in provider_supported_products]
            raise UnsupportedProvider("The requested provider is not (yet) supported")
        return [dict(
            ID=code,
            **value
        ) for code, value in self.product_types_config.items()]

    def available_providers(self):
        """Gives the list of the available providers"""
        return [provider for provider in self.providers_config]

    def search(self, product_type, **kwargs):
        """Look for products matching criteria in known systems.

        :param product_type: The product type to search
        :type product_type: str or unicode
        :param dict kwargs: some other criteria that will be used to do the search
        :returns: A collection of EO products matching the criteria
        :rtype: :class:`~eodag.api.search_result.SearchResult`

        .. note::
            The search interfaces, which are implemented as plugins, are required to return a list as a result of their
            processing. This requirement is enforced here.
        """
        search_plugins = self._build_search_plugins(product_type)
        results = SearchResult([])
        for idx, plugin in enumerate(search_plugins):
            logger.debug('Using plugin for search: %s on provider: %s', plugin.name, plugin.instance_name)
            auth = self._build_auth_plugin(plugin.instance_name)
            try:
                r = plugin.query(product_type, auth=auth, **kwargs)
                if not isinstance(r, list):
                    raise PluginImplementationError(
                        'The query function of a Search plugin must return a list of results, got {} '
                        'instead'.format(type(r)))
                # Filter and attach to each eoproduct in the result the plugin capable of downloading it (this
                # is done to enable the eo_product to download itself doing: eo_product.download())
                # The filtering is done by keeping only those eo_products that intersects the search extent (if there
                # was no search extent, search_intersection contains the geometry of the eo_product)
                # WARNING: this means an eo_product that has an invalid geometry can still be returned as a search
                # result if there was no search extent (because we will not try to do an intersection)
                for eo_product in r:
                    if eo_product.search_intersection is not None:
                        if 'download' in self.providers_config[eo_product.provider]:
                            topic = 'download'
                        else:
                            topic = 'api'
                        eo_product.register_downloader(
                            self._build_provider_plugin(eo_product.provider, topic),
                            auth
                        )
                        results.append(eo_product)
                # Decide if we should go on with the search using the other search plugins. This happens in 2 cases:
                # 1. The currently used plugin is the preferred one (the first), and it returned no result
                # 2. The currently used plugin supports the product_type partially
                if not plugin.config.get('products', {}).get(product_type, {}).get('partial', False):
                    if idx == 0 and len(search_plugins) > 1 and len(r) == 0:
                        logger.debug(
                            "No result from preferred provider: '%r'. Search continues on other providers "
                            "supporting the product type: '%r'", plugin.instance_name, product_type)
                        continue
                    break
                logger.debug("Detected partial product type '%s' on priviledged instance '%s'. Search continues on "
                             "other instances supporting it.", product_type, plugin.instance_name)
            except RuntimeError as rte:
                if 'Unknown product type' in rte.args:
                    logger.debug('Product type %s not known by %s instance', product_type, plugin.instance_name)
                else:
                    raise rte
            except Exception:
                import traceback as tb
                logger.debug('Error while searching on interface %s:\n %s.', plugin, tb.format_exc())
                logger.debug('Ignoring it')
        return results

    @staticmethod
    def sort_by_extent(searches):
        """Combines multiple SearchResults and return a list of SearchResults sorted by extent.

        :param searches: List of eodag SearchResult
        :type searches: list
        :return: list of :class:`~eodag.api.search_result.SearchResult`
        """
        products_grouped_by_extent = {}

        for search in searches:
            for product in search:
                same_geom = products_grouped_by_extent.setdefault(
                    ''.join([str(round(p, 2)) for p in product.geometry.bounds]), []
                )
                same_geom.append(product)

        return [
            SearchResult(products_grouped_by_extent[extent_as_wkb_hex])
            for extent_as_wkb_hex in products_grouped_by_extent
        ]

    def download_all(self, search_result, progress_callback=None):
        """Download all products resulting from a search.

        :param search_result: A collection of EO products resulting from a search
        :type search_result: :class:`~eodag.api.search_result.SearchResult`
        :param progress_callback: (optional) A method or a callable object
                                  which takes a current size and a maximum
                                  size as inputs and handle progress bar
                                  creation and update to give the user a
                                  feedback on the download progress
        :type progress_callback: :class:`~eodag.utils.ProgressCallback` or None
        :returns: A collection of the absolute paths to the downloaded products
        :rtype: list
        """
        paths = []
        if search_result:
            with tqdm(search_result, unit='product', desc='Downloading products') as bar:
                for product in bar:
                    paths.append(self.download(product, progress_callback=progress_callback))
        else:
            print('Empty search result, nothing to be downloaded !')
        return paths

    @staticmethod
    def serialize(search_result, filename='search_results.geojson'):
        """Registers results of a search into a geojson file.

        :param search_result: A collection of EO products resulting from a search
        :type search_result: :class:`~eodag.api.search_result.SearchResult`
        :param filename: The name of the file to generate
        :type filename: str or unicode
        :returns: The name of the created file
        :rtype: str or unicode
        """
        with open(filename, 'w') as fh:
            geojson.dump(search_result, fh)
        return filename

    @staticmethod
    def deserialize(filename):
        """Loads results of a search from a geojson file.

        :param filename: A filename containing a search result encoded as a geojson
        :type filename: str or unicode
        :returns: The search results encoded in `filename`
        :rtype: :class:`~eodag.api.search_result.SearchResult`
        """
        with open(filename, 'r') as fh:
            return SearchResult.from_geojson(geojson.load(fh))

    def download(self, product, progress_callback=None):
        """Download a single product.

        .. warning::
            As a side effect, this method changes the `location` attribute of an :class:`~eodag.api.product.EOProduct`.
            So be aware that you may not be able to download the product again if you didn't store its remote location
            somewhere first before downloading it, unless the download failed or the download plugin didn't returned
            the local path of the downloaded resource

        :param product: The EO product to download
        :type product: :class:`~eodag.api.product.EOProduct`
        :param progress_callback: (optional) A method or a callable object
                                  which takes a current size and a maximum
                                  size as inputs and handle progress bar
                                  creation and update to give the user a
                                  feedback on the download progress
        :type progress_callback: :class:`~eodag.utils.ProgressCallback` or None
        :returns: The absolute path to the downloaded product in the local filesystem
        :rtype: str or unicode
        :raises: :class:`~eodag.utils.exceptions.PluginImplementationError`
        :raises: :class:`RuntimeError`
        """
        if progress_callback is None:
            progress_callback = ProgressCallback()

        download_plugin = self._build_download_plugin(product)
        logger.debug('Using download plugin: %s for provider: %s', download_plugin.name, download_plugin.instance_name)
        try:
            auth = None
            # We need to authenticate only if we are not on the platform where the product lives
            if not download_plugin.config.get('on_site', False):
                auth = self._build_auth_plugin(download_plugin.instance_name)
            else:
                logger.debug('On site usage detected. Authentication for download skipped !')
            if auth:
                auth = auth.authenticate()
            product_location = download_plugin.download(product, auth=auth, progress_callback=progress_callback)
            if product_location is None:
                logger.warning('The download method of a Download plugin should return the absolute path to the '
                               'downloaded resource')
            else:
                logger.debug('Product location updated from %s to %s', product.location, product_location)
                # Update the product location if and only if a path was returned by the download plugin
                product.location = 'file://{}'.format(product_location)
            return product_location
        except TypeError as e:
            # Enforcing the requirement for download plugins to implement a download method with auth kwarg
            if any("got an unexpected keyword argument 'auth'" in arg for arg in e):
                raise PluginImplementationError('The download method of a Download plugin must support auth keyword '
                                                'argument')
            raise e
        except RuntimeError as rte:
            # Skip download errors, allowing other downloads to take place anyway
            if 'is incompatible with download plugin' in rte:
                logger.warning('Download plugin incompatibility found. Skipping download...')
            else:
                raise rte

    def _build_auth_plugin(self, provider_name):
        if 'auth' in self.providers_config[provider_name]:
            logger.debug('Authentication initialisation for provider: %s', provider_name)
            plugin = self._build_provider_plugin(provider_name, 'auth')
            logger.debug("Initialized %r Authentication plugin for provider: '%s'", plugin, provider_name)
            return plugin

    def _build_search_plugins(self, product_type):
        """Look for search interfaces to use, based on the configuration of the api"""
        logger.debug('Looking for the appropriate Search instance(s) to use for product type: %s', product_type)
        previous = self._plugins_cache.setdefault('search', {}).setdefault(product_type, [])
        if not previous:
            search_plugins = self._plugins_manager.instantiate_configured_plugins(
                topics=('search', 'api'),
                product_type_id=product_type
            )
            # Store the newly instantiated plugins in the cache
            previous.extend(search_plugins)
        # The searcher used will be the one with highest priority
        previous.sort(key=attrgetter('priority'), reverse=True)
        logger.debug("Found %s Search instance(s) for product type '%s' (ordered by highest priority): %r",
                     len(previous), product_type, previous)
        return previous

    def _build_download_plugin(self, product):
        """Look for a download plugin to use, based on the configuration of the api and the product to download"""
        logger.debug('Looking for the appropriate Download plugin to use for product: %r', product)
        if 'download' in self.providers_config[product.provider]:
            topic = 'download'
        else:
            topic = 'api'
        plugin = self._build_provider_plugin(product.provider, topic)
        logger.debug('Found Download plugin for product %r: %s', product, plugin)
        return plugin

    def _build_provider_plugin(self, provider, topic):
        """Build the plugin of type topic for provider

        :param provider: The provider for which to build topic plugin
        :type provider: str or unicode
        :param topic: The type of plugin to build for provider (one of: 'auth', 'search', 'download'). Note that this
                      method is not intended to be used for 'crunch' plugins, because these are generally not tied to a
                      specific provider
        :type topic: str or unicode
        :returns: The plugin instance corresponding to the topic for the specified provider
        :rtype: a :class:`~eodag.plugins.base.PluginTopic` subclass's instance
        """
        previous = self._plugins_cache.setdefault(topic, {}).setdefault(provider, None)
        if previous is None:
            previous = self._plugins_manager.instantiate_plugin_by_config(
                topic_name=topic,
                topic_config=self.providers_config[provider][topic],
                iname=provider,
            )
        return previous

    def get_cruncher(self, name, **options):
        plugin_conf = {
            'plugin': name,
        }
        plugin_conf.update({
            key.replace('-', '_'): val
            for key, val in options.items()
        })
        return self._plugins_manager.instantiate_plugin_by_config('crunch', plugin_conf)



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import logging

from eodag.api.product import EOProduct
from eodag.api.product.representations import properties_from_json
from eodag.plugins.search.resto import RestoSearch


logger = logging.getLogger('eodag.plugins.search.aws')


class AwsSearch(RestoSearch):

    def __init__(self, config):
        super(AwsSearch, self).__init__(config)
        self.dl_base = config.get('amazon_dl_endpoint')

    def normalize_results(self, product_type, results, search_bbox):
        normalized = []
        if results['features']:
            logger.info('Found %s products', len(results['features']))
            logger.debug('Adapting plugin results to eodag product representation')
            for result in results['features']:
                ref = result['properties']['title'].split('_')[5]
                year = result['properties']['completionDate'][0:4]
                month = str(int(result['properties']['completionDate'][5:7]))
                day = str(int(result['properties']['completionDate'][8:10]))

                download_url = ('{proto}://tiles/{ref[1]}{ref[2]}/{ref[3]}/{ref[4]}{ref[5]}/{year}/'
                                '{month}/{day}/0/').format(proto=self.config['product_location_scheme'], **locals())

                product = EOProduct(
                    product_type,
                    self.instance_name,
                    download_url,
                    properties_from_json(result, self.config['metadata_mapping']),
                    searched_bbox=search_bbox,
                )
                normalized.append(product)
            logger.debug('Normalized products : %s', normalized)
        else:
            logger.info('Nothing found !')
        return normalized



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

import re
import sys
import types
import unicodedata
from datetime import datetime
from itertools import repeat, starmap

import click
import pyproj
from requests.auth import AuthBase
from tqdm import tqdm


# All modules using these should import them from utils package
try:  # PY3
    from urllib.parse import urljoin, urlparse, parse_qs, urlunparse  # noqa
except ImportError:  # PY2
    from urlparse import urljoin, urlparse, parse_qs, urlunparse  # noqa

try:    # PY3
    from urllib.parse import urlencode  # noqa
except ImportError:     # PY2
    from urllib import urlencode    # noqa


class RequestsTokenAuth(AuthBase):
    token = ''

    def __call__(self, req):
        req.headers['Authorization'] = "Bearer {}".format(self.token)
        return req


class RequestsDictTokenAuth(RequestsTokenAuth):
    def __init__(self, token_obj, token_key):
        self.token = token_obj[token_key]


class RequestsTextTokenAuth(RequestsTokenAuth):
    def __init__(self, token):
        self.token = token


class FloatRange(click.types.FloatParamType):
    """A parameter that works similar to :data:`click.FLOAT` but restricts the value to fit into a range. Fails if the
    value doesn't fit into the range.
    """
    name = 'percentage'

    def __init__(self, min=None, max=None):
        self.min = min
        self.max = max

    def convert(self, value, param, ctx):
        rv = click.types.FloatParamType.convert(self, value, param, ctx)
        if self.min is not None and rv < self.min or self.max is not None and rv > self.max:
            if self.min is None:
                self.fail('%s is bigger than the maximum valid value '
                          '%s.' % (rv, self.max), param, ctx)
            elif self.max is None:
                self.fail('%s is smaller than the minimum valid value '
                          '%s.' % (rv, self.min), param, ctx)
            else:
                self.fail('%s is not in the valid range of %s to %s.'
                          % (rv, self.min, self.max), param, ctx)
        return rv

    def __repr__(self):
        return 'FloatRange(%r, %r)' % (self.min, self.max)


def slugify(value, allow_unicode=False):
    """Copied from Django Source code, only modifying last line (no need for safe strings).
    source: https://github.com/django/django/blob/master/django/utils/text.py

    Convert to ASCII if 'allow_unicode' is False. Convert spaces to hyphens.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Convert to lowercase. Also strip leading and trailing whitespace.
    """
    try:  # PY2
        value = unicode(value)
    except NameError:  # PY3
        value = str(value)
    if allow_unicode:
        value = unicodedata.normalize('NFKC', value)
    else:
        value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
    value = re.sub(r'[^\w\s-]', '', value).strip().lower()
    return re.sub(r'[-\s]+', '-', value)


def utf8_everywhere(mapping):
    """Recursively transforms all string found in the dict mapping to UTF-8 if we are on Python 2"""
    mutate_dict_in_place((
        lambda value:
        value.decode('utf-8')
        if isinstance(value, str) and sys.version_info.major == 2 and sys.version_info.minor == 7
        else value),
        mapping
    )


def mutate_dict_in_place(func, mapping):
    """Apply func to values of mapping.

    The mapping object's values are modified in-place. The function is recursive, allowing to also modify values of
    nested dicts that may be level-1 values of mapping.

    :param func: A function to apply to each value of mapping which is not a dict object
    :type func: func
    :param mapping: A Python dict object
    :type mapping: dict
    :returns: None
    """
    for key, value in mapping.items():
        if isinstance(value, dict):
            mutate_dict_in_place(func, value)
        else:
            mapping[key] = func(value)


def maybe_generator(obj):
    """Generator function that get an arbitrary object and generate values from it if the object is a generator."""
    if isinstance(obj, types.GeneratorType):
        for elt in obj:
            yield elt
    else:
        yield obj


DEFAULT_PROJ = pyproj.Proj(init='EPSG:4326')


def get_timestamp(date_time, date_format='%Y-%m-%d'):
    """Returns the given date_time string formatted with date_format as timestamp, in a PY2/3 compatible way

    :param date_time: the datetime string to return as timestamp
    :type date_time: str or unicode
    :param date_format: (optional) the date format in which date_time is given, defaults to '%Y-%m-%d'
    :type date_format: str or unicode
    :returns: the timestamp corresponding to the date_time string in seconds
    :rtype: float
    """
    date_time = datetime.strptime(date_time, date_format)
    try:
        return date_time.timestamp()
    except AttributeError:  # There is no timestamp method on datetime objects in Python 2
        import time
        return time.mktime(date_time.timetuple()) + date_time.microsecond / 1e6


class ProgressCallback(object):
    """A callable used to render progress to users for long running processes"""

    def __init__(self):
        self.pb = None

    def __call__(self, current_size, max_size):
        """Update the progress bar.

        :param current_size: amount of data already processed
        :type current_size: int
        :param max_size: maximum amount of data to be processed
        :type max_size: int
        """
        if self.pb is None:
            self.pb = tqdm(total=max_size, unit='KB', unit_scale=True)
        self.pb.update(current_size)


def repeatfunc(func, n, *args):
    """Call `func` `n` times with `args`"""
    return starmap(func, repeat(args, n))



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import getpass

import requests


def get_airbus_ds_dias_product_types():
    """Get all the collections and product types that are supported by airbus dias provider (an
    `ARLAS server <http://arlas.io/arlas-tech/current/>`_ implemented with a collection named `catalog`, located at
    https://testing.dias.datastore.multisat.space/sdk/resolver/api/v1/services/explore/explore/catalog/
    """
    base = 'https://testing.dias.datastore.multisat.space/sdk10/lobby/api/v1/services/explore/explore/catalog/_search'
    query = 'f=identification.collection:like:Sentinel&include=identification.collection,identification.type'
    url = '?'.join(['{base}', '{query}']).format(**locals())
    try:
        user = raw_input('Airbus DIAS user: ')
    except NameError:
        user = input('Airbus DIAS user: ')
    password = getpass.getpass("Airbus DIAS password for user '{}':".format(user))
    result = requests.get(url, auth=(user, password)).json()
    if result['nbhits'] < result['totalnb']:
        rest = requests.get(
            url + '&from={r[nbhits]}&size={r[totalnb]}'.format(r=result),
            auth=(user, password)
        ).json()
        result['hits'].extend(rest['hits'])
    aggregated_result = {}
    for product_identification in (hit['data']['identification'] for hit in result['hits']):
        aggregated_result.setdefault(
            product_identification['collection'],
            set()
        ).add(product_identification['type'])
    for collection, product_types in aggregated_result.items():
        yield collection, product_types



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import os

import yaml
import yaml.parser

from eodag.utils import utf8_everywhere


class SimpleYamlProxyConfig(object):
    """A simple configuration class acting as a proxy to an underlying dict object as returned by yaml.load"""

    def __init__(self, conf_file_path):
        with open(os.path.abspath(os.path.realpath(conf_file_path)), 'r') as fh:
            try:
                self.source = yaml.load(fh)
                utf8_everywhere(self.source)
            except yaml.parser.ParserError as e:
                print('Unable to load user configuration file')
                raise e

    def __getitem__(self, item):
        return self.source[item]

    def __contains__(self, item):
        return item in self.source

    def __iter__(self):
        return iter(self.source)

    def items(self):
        return self.source.items()

    def values(self):
        return self.source.values()

    def update(self, other):
        if not isinstance(other, self.__class__):
            raise ValueError("'{}' must be of type {}".format(other, self.__class__))
        self.source.update(other.source)



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import zipfile

import numpy
import rasterio
import requests
import xarray as xr
from rasterio.enums import Resampling
from rasterio.vrt import WarpedVRT
from requests import HTTPError
from shapely import geometry
from tqdm import tqdm

from eodag.utils import ProgressCallback


try:
    from shapely.errors import TopologicalError
except ImportError:
    from shapely.geos import TopologicalError

from eodag.api.product.drivers import DRIVERS, NoDriver
from eodag.api.product.representations import DEFAULT_METADATA_MAPPING, properties_from_json
from eodag.utils.exceptions import UnsupportedDatasetAddressScheme


logger = logging.getLogger('eodag.api.product')


class EOProduct(object):
    """A wrapper around an Earth Observation Product originating from a search.

    Every Search plugin instance must build an instance of this class for each of the result of its query method, and
    return a list of such instances.

    :param product_type: The product type of the product as defined in eodag
    :type product_type: str or unicode
    :param provider: The provider from which the product originates
    :type provider: str or unicode
    :param download_url: A uri informing where to go to download the product
    :type download_url: str or unicode
    :param properties: The metadata of the product
    :type properties: dict
    :param searched_bbox: (optional) The extent that was passed as a search constraint, used to attach to the EOProduct
                           the intersection of its geometry with this extent
    :type searched_bbox: dict

    .. note::
        The geojson spec `enforces <https://github.com/geojson/draft-geojson/pull/6>`_ the expression of geometries as
        WGS84 CRS (EPSG:4326) coordinates and EOProduct is intended to be transmitted as geojson between applications.
        Therefore it stores geometries in the before mentioned CRS.
    """

    def __init__(self, product_type, provider, download_url, properties, searched_bbox=None):
        self.product_type = product_type
        self.provider = provider
        self.location = download_url
        self.properties = properties
        self.geometry = self.search_intersection = geometry.shape(self.properties['geometry'])
        if searched_bbox is not None:
            searched_bbox_as_shape = geometry.box(searched_bbox['lonmin'], searched_bbox['latmin'],
                                                  searched_bbox['lonmax'], searched_bbox['latmax'])
            try:
                self.search_intersection = self.geometry.intersection(searched_bbox_as_shape)
            except TopologicalError:
                import traceback as tb
                logger.warning('Unable to intersect the requested extent: %s with the product geometry: %s. Got:\n%s',
                               searched_bbox_as_shape, self.properties['geometry'], tb.format_exc())
                self.search_intersection = None
        self.driver = DRIVERS.get(self.product_type, NoDriver())
        self.downloader = None
        self.downloader_auth = None

    def get_data(self, crs, resolution, band, extent):
        """Retrieves all or part of the raster data abstracted by the :class:`EOProduct`

        :param crs: The coordinate reference system in which the dataset should be returned
        :type crs: str or unicode
        :param resolution: The resolution in which the dataset should be returned (given in the unit of the crs)
        :type resolution: float
        :param band: The band of the dataset to retrieve (e.g.: 'B01')
        :type band: str or unicode
        :param extent: The coordinates on which to zoom as a tuple (min_x, min_y, max_x, max_y) in the given `crs`
        :type extent: (float, float, float, float)
        :returns: The numeric matrix corresponding to the sub dataset or an empty array if unable to get the data
        :rtype: xarray.DataArray
        """
        fail_value = xr.DataArray(numpy.empty(0))
        try:
            dataset_address = self.driver.get_data_address(self, band)
        except UnsupportedDatasetAddressScheme:
            logger.warning('Eodag does not support getting data from distant sources by now. Falling back to first '
                           'downloading the product and then getting the data...')
            try:
                path_of_downloaded_file = self.download()
            except RuntimeError:
                import traceback
                logger.warning('Error while trying to download the product:\n %s', traceback.format_exc())
                logger.warning('There might be no download plugin registered for this EO product. Try performing: '
                               'product.register_downloader(download_plugin, auth_plugin) before trying to call '
                               'product.get_data(...)')
                return fail_value
            if not path_of_downloaded_file:
                return fail_value
            dataset_address = self.driver.get_data_address(self, band)
        min_x, min_y, max_x, max_y = extent
        height = int((max_y - min_y) / resolution)
        width = int((max_x - min_x) / resolution)
        out_shape = (width, height)
        with rasterio.open(dataset_address) as src:
            with WarpedVRT(src, dst_crs=crs, resampling=Resampling.bilinear) as vrt:
                array = vrt.read(1, window=vrt.window(*extent), out_shape=out_shape, resampling=Resampling.bilinear)
                return xr.DataArray(array)

    def as_dict(self):
        """Builds a representation of EOProduct as a dictionary to enable its geojson serialization

        :returns: The representation of a :class:`~eodag.api.product.EOProduct` as a Python dict
        :rtype: dict
        """
        geojson_repr = {
            'type': 'Feature',
            'geometry': geometry.mapping(self.geometry),
            'id': self.properties['id'],
            'properties': {
                'eodag_product_type': self.product_type,
                'eodag_provider': self.provider,
                'eodag_download_url': self.location,
                'eodag_search_intersection': geometry.mapping(self.search_intersection),
            }
        }
        geojson_repr['properties'].update({
            key: value
            for key, value in self.properties.items()
            if key not in ('geometry', 'id')
        })
        return geojson_repr

    @classmethod
    def from_geojson(cls, feature):
        """Builds an :class:`~eodag.api.product.EOProduct` object from its representation as geojson

        :param feature: The representation of a :class:`~eodag.api.product.EOProduct` as a Python dict
        :type feature: dict
        :returns: An instance of :class:`~eodag.api.product.EOProduct`
        :rtype: :class:`~eodag.api.product.EOProduct`
        """
        obj = cls(
            feature['properties']['eodag_product_type'],
            feature['properties']['eodag_provider'],
            feature['properties']['eodag_download_url'],
            properties_from_json(feature, DEFAULT_METADATA_MAPPING)
        )
        obj.search_intersection = feature['properties']['eodag_search_intersection']
        return obj

    # Implementation of geo-interface protocol (See https://gist.github.com/sgillies/2217756)
    __geo_interface__ = property(as_dict)

    def __repr__(self):
        return '{}(id={}, provider={})'.format(self.__class__.__name__, self.properties['id'], self.provider)

    def encode(self, raster, encoding='protobuf'):
        """Encode the subset to a network-compatible format.

        :param raster: The raster data to encode
        :type raster: xarray.DataArray
        :param encoding: The encoding of the export
        :type encoding: str or unicode
        :return: The data encoded in the specified encoding
        :rtype: bytes
        """
        # If no encoding return an empty byte
        if not encoding:
            logger.warning('Trying to encode a raster without specifying an encoding')
            return b''
        strategy = getattr(self, '_{encoding}'.format(**locals()), None)
        if strategy:
            return strategy(raster)
        logger.error('Unknown encoding: %s', encoding)
        return b''

    def _protobuf(self, raster):
        """Google's Protocol buffers encoding strategy.

        :param raster: The raster to encode
        :type raster: xarray.DataArray
        :returns: The raster data represented by this subset in protocol buffers encoding
        :rtype: bytes
        """
        from eodag.api.product.protobuf import eo_product_pb2
        subdataset = eo_product_pb2.EOProductSubdataset()
        subdataset.id = self.properties['id']
        subdataset.producer = self.provider
        subdataset.product_type = self.product_type
        subdataset.platform = self.properties['platformSerialIdentifier']
        subdataset.sensor = self.properties['instrument']
        data = subdataset.data
        data.array.extend(list(raster.values.flatten().astype(int)))
        data.shape.extend(list(raster.values.shape))
        data.dtype = raster.values.dtype.name
        return subdataset.SerializeToString()

    def register_downloader(self, downloader, authenticator):
        """Give to the product the information needed to download itself.

        :param downloader: The download method that it can use
        :type downloader: Concrete subclass of :class:`~eodag.plugins.download.base.Download` or
                          :class:`~eodag.plugins.api.base.Api`
        :param authenticator: The authentication method needed to perform the download
        :type authenticator: Concrete subclass of :class:`~eodag.plugins.authentication.base.Authentication`
        """
        self.downloader = downloader
        self.downloader_auth = authenticator

    def download(self, progress_callback=None):
        """Download the EO product using the provided download plugin and the authenticator if necessary.

        The actual download of the product occurs only at the first call of this method. A side effect of this method
        is that it changes the `location` attribute of an EOProduct, from its remote address to the local address.

        :param progress_callback: (optional) A method or a callable object
                                  which takes a current size and a maximum
                                  size as inputs and handle progress bar
                                  creation and update to give the user a
                                  feedback on the download progress
        :type progress_callback: :class:`~eodag.utils.ProgressCallback` or None
        :returns: The absolute path to the downloaded product on the local filesystem
        :rtype: str or unicode
        """
        if progress_callback is None:
            progress_callback = ProgressCallback()

        if self.downloader is None:
            raise RuntimeError('EO product is unable to download itself due to the lack of a download plugin')
        # Remove the capability for the downloader to perform extraction if the downloaded product is a zipfile. This
        # way, the eoproduct is able to control how it stores itself on the local filesystem
        old_extraction_config = self.downloader.config['extract']
        self.downloader.config['extract'] = False
        auth = self.downloader_auth.authenticate() if self.downloader_auth is not None else self.downloader_auth
        fs_location = self.downloader.download(self, auth=auth, progress_callback=progress_callback)
        if fs_location is None:
            logger.warning('The download may have fail or the location of the downloaded file on the local filesystem '
                           'have not been returned by the download plugin')
            return ''
        if zipfile.is_zipfile(fs_location):
            # Unzip only if it was not done before
            if not os.path.exists(fs_location[:fs_location.index('.zip')]):
                with zipfile.ZipFile(fs_location, 'r') as zfile:
                    fileinfos = tqdm(zfile.infolist(), unit='file',
                                     desc='Extracting files from {}'.format(fs_location))
                    for fileinfo in fileinfos:
                        zfile.extract(fileinfo, path=fs_location[:fs_location.index('.zip')])
            # Handle depth levels in the product archive. For example, if the downloaded archive was
            # extracted to: /top_level/product_base_dir and archive_depth was configured to 2, the product
            # location will be /top_level/product_base_dir.
            # WARNING: A strong assumption is made here: there is only one subdirectory per level
            archive_depth = self.downloader.config.get('archive_depth', 1)
            fs_location = fs_location[:fs_location.index('.zip')]
            count = 1
            while count < archive_depth:
                fs_location = os.path.join(fs_location, os.listdir(fs_location)[0])
                count += 1
        # After the product has been downloaded, we need to modify its location attribute to reflect that it is now
        # in the filesystem
        logger.debug('Product location updated from %s to %s', self.location, fs_location)
        self.location = 'file://{}'.format(fs_location)
        # Restore configuration
        self.downloader.config['extract'] = old_extraction_config
        return fs_location

    def get_quicklook(self, filename=None, progress_callback=None):
        """Download the quick look image of a given EOProduct from its provider if it exists.

        :param filename: (optional) the name to give to the downloaded quicklook.
        :type filename: str (Python 3) or unicode (Python 2)
        :param progress_callback: (optional) A method or a callable object
                                  which takes a current size and a maximum
                                  size as inputs and handle progress bar
                                  creation and update to give the user a
                                  feedback on the download progress
        :type progress_callback: :class:`~eodag.utils.ProgressCallback` or None
        :returns: The absolute path of the downloaded quicklook
        :rtype: str (Python 3) or unicode (Python 2)
        """
        if progress_callback is None:
            progress_callback = ProgressCallback()

        if self.properties['quicklook'] is None:
            logger.warning('Missing information to retrieve quicklook for EO product: %s', self.properties['id'])
            return ''

        quicklooks_base_dir = os.path.join(self.downloader.config['outputs_prefix'], "quicklooks")
        if not os.path.isdir(quicklooks_base_dir):
            os.makedirs(quicklooks_base_dir)
        quicklook_file = os.path.join(
            quicklooks_base_dir,
            filename
            if filename is not None
            else self.properties['id']
        )

        if not os.path.isfile(quicklook_file):
            auth = self.downloader_auth.authenticate() if self.downloader_auth is not None else None
            with requests.get(self.properties['quicklook'], stream=True, auth=auth) as stream:
                try:
                    stream.raise_for_status()
                except HTTPError:
                    import traceback as tb
                    logger.error("Error while getting resource :\n%s", tb.format_exc())
                    return ''
                else:
                    stream_size = int(stream.headers.get('content-length', 0))
                    with open(quicklook_file, 'wb') as fhandle:
                        for chunk in stream.iter_content(chunk_size=64 * 1024):
                            if chunk:
                                fhandle.write(chunk)
                                progress_callback(len(chunk), stream_size)
                    logger.info('Download recorded in %s', quicklook_file)
        return quicklook_file



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals


def get_metadata_path(map_value):
    """Return the jsonpath or xpath to the value of a EO product metadata in a provider search result.

    The path is retrieved depending on if the metadata is queryable (the value associated to it in the provider search
    config metadata mapping is a list) or not (the value is directly the string corresponding to the path).

    Assume we have the following provider config::

        provider:
            ...
            search:
                ...
                metadata_mapping:
                    productType:
                        - productType
                        - $.properties.productType
                    id: $.properties.id
                    ...
                ...
            ...

    Then the metadata `id` is not queryable for this provider meanwhile `productType` is queryable. The first value of
    the `metadata_mapping.productType` is how the eodag search parameter `productType` is interpreted in the
    :class:`~eodag.plugins.search.base.Search` plugin implemented by `provider`, and is used when eodag delegates
    search process to the corresponding plugin.

    :param map_value: The value originating from the definition of `metadata_mapping` in the provider search config.
        For example, it is the list `['productType', '$.properties.productType']` with the sample above
    :type map_value: str (Python 3) or unicode (Python 2) or list(str or unicode)
    :return: The value of the path to the metadata value in the provider search result
    :rtype: str (Python 3) or unicode (Python 2)
    """
    if isinstance(map_value, list):
        path = map_value[1]
    else:
        path = map_value
    return path


def get_search_param(map_value):
    """See :func:`~eodag.utils.metadata_mapping.get_metadata_path`

    :param map_value: The value originating from the definition of `metadata_mapping` in the provider search config
    :type map_value: list
    :return: The value of the search parameter as defined in the provider config
    :rtype: str (Python 3) or unicode (Python 2)
    """
    # Assume that caller will pass in the value as a list
    return map_value[0]



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import logging
from collections import Iterable

import pkg_resources

from eodag.plugins.apis.base import Api
from eodag.plugins.authentication.base import Authentication
from eodag.plugins.base import GeoProductDownloaderPluginMount
from eodag.plugins.crunch.base import Crunch
from eodag.plugins.download.base import Download
from eodag.plugins.search.base import Search


logger = logging.getLogger('eodag.plugins.instances_manager')


class PluginInstancesManager(object):
    """A manager for the plugins instances.

    The role of instances of this class (normally only one instance exists, created during instantiation of
    :class:`~eodag.api.core.EODataAccessGateway`, and referenced by :attr:`eodag.api.core.EODataAccessGateway.pim` -
    pim stands for plugin instances manager. But nothing is done to enforce this) is to instantiate the plugins
    according to the providers configuration, and keep track of them in memory. The providers configuration contains
    information such as the name of the provider, the internet endpoint for accessing it, and the plugins to use to
    perform defined actions (search, download, authenticate, crunch).

    If the providers configuration looks like this::

        conf = {
            'theia': {
                'search': {
                    'plugin': 'RestoSearch',
                    'api_endpoint': ...,
                    ...
                },
                'download': {
                    'plugin': 'HTTPDownload',
                    ...
                },
                'crunch': {
                    'plugin': 'VeryComplexCruncher',
                    ...
                }
            },
            ...
        }

    the manager will ask to the :class:`~eodag.plugins.search.base.Search` class to give it back the
    :class:`~eodag.plugins.search.resto.RestoSearch` plugin class by calling
    ``Search.get_plugin_by_name(conf['theia']['search']['plugin'])``, so that it can create its ``theia`` instance
    with ``conf['theia']['search']`` configuration.

    :param providers_config: The configuration with all information about the providers supported by the `eodag`
    :type providers_config: :class:`~eodag.config.SimpleYamlProxyConfig`
    """
    supported_topics = {
        'search': Search,
        'download': Download,
        'crunch': Crunch,
        'auth': Authentication,
        'api': Api,
    }

    def __init__(self, providers_config):
        self.providers_config = providers_config
        # Load all the plugins. This will make all plugin classes of a particular type to be available in the base
        # plugin class's 'plugins' attribute. For example, by importing module 'eodag.plugins.search.resto', the plugin
        # 'RestoSearch' will be available in self.supported_topics['search'].plugins
        for topic in self.supported_topics:
            # This way of discovering plugins means that anyone can create eodag plugins as a separate python package
            # (though it must require eodag), and have it discovered as long as they declare an entry point of the type
            # 'eodag.plugins.search' for example in its setup script. See the setup script of eodag for an example of
            # how to do this.
            for entry_point in pkg_resources.iter_entry_points('eodag.plugins.{}'.format(topic)):
                try:
                    entry_point.load()
                except ImportError:
                    import traceback as tb
                    logger.warning('Unable to load plugin: %s.', entry_point.name)
                    logger.warning('Reason:\n%s', tb.format_exc())
                    logger.warning('Check that the plugin module (%s) is importable', entry_point.module_name)

    def instantiate_configured_plugins(self, topics, product_type_id='', providers=None):
        """Instantiate all known plugins of particular type.

        :param topics: An iterable of plugin topics (e.g: ``('search', 'download')``).
        :type topics: :class:`~collections.Iterable(str)`
        :param product_type_id: (optional) A product type ID, that will help to filter which instances to create (the
                                ones which explicitly support the product type as can be seen in the providers config).
        :type product_type_id: str or unicode
        :param providers: (optional) An iterable of tuples (provider_name, priority) to enable instantiating only these
                          providers with the given priority (they should be still configured for this to work).
        :type providers: Iterable(tuple(str, int))
        :returns: A list of instantiated and configured plugins of the given topics and supporting the given product
                  type, or only the plugins corresponding to the topics for the providers specified in ``providers``.
        :rtype: list
        :raises: :class:`AssertionError` if neither ``product_type_id`` nor ``providers`` keyword arguments is given.
        """
        assert any((product_type_id, providers)), ("You should provide at least one of 'product_type_id' or "
                                                   "'providers' parameters")
        if isinstance(topics, Iterable):
            instances = []
            for topic in topics:
                instances.extend(
                    self._instantiate(topic, providers=self._filter_providers(topic, product_type_id, providers)))
        else:
            instances = self._instantiate(topics, providers=self._filter_providers(topics, product_type_id, providers))
        return instances

    def instantiate_plugin_by_config(self, topic_name, topic_config, base=None, iname='', priority=None):
        """Creates an instance of a type of plugin based on a configuration suited for this type of plugin.

        :param topic_name: The name of the type of plugin to instantiate (e.g: ``'search'``)
        :type topic_name: str or unicode
        :param dict topic_config: A configuration compatible for the type of plugin to instantiate
        :param base: (optional) The python class corresponding to the `topic_name`
        :type base: subclass of :class:`~eodag.plugins.base.PluginTopic` or None
        :param str iname: (optional) The name to give to the created instance
        :param priority: The priority that will be given to the created instance
        :type priority: int or None
        :returns: An instance of a subclass of `base` or the class corresponding to `topic_name`
        """
        logger.debug("Creating '%s' plugin instance with config '%s'", topic_name.upper(),
                     {key: value for key, value in topic_config.items() if key != 'credentials'})
        PluginBaseClass = base or self._get_base_class(topic_name)
        PluginClass = GeoProductDownloaderPluginMount.get_plugin_by_name(
            PluginBaseClass,
            topic_config['plugin']
        )
        instance = PluginClass(topic_config)
        # Each instance will be identified by its name, independently of the plugin implementation
        instance.instance_name = iname
        # Update instance priority
        if priority and priority != instance.priority:
            instance.priority = priority
        return instance

    def instantiate_plugin_by_name(self, topic, name):
        """Creates an instance of a type of plugin based on its name (the name of the Python class of the plugin).

        This method is intended to be used for 'config-free' plugins (e.g. Crunch plugins). These are plugins that do
        not need configuration to be instantiated.

        :param topic: The type of plugin to be instantiated (e.g.: 'crunch')
        :type topic: str or unicode
        :param name: The name of the Python class representing the plugin
        :type name: str or unicode
        :returns: An instance of the class corresponding to `name`
        """
        logger.debug("Creating '%s' plugin instance with name '%s' (config-free instance)", topic.upper(), name)
        PluginBaseClass = self._get_base_class(topic)
        PluginClass = GeoProductDownloaderPluginMount.get_plugin_by_name(PluginBaseClass, name)
        return PluginClass()

    def _get_topics(self, provider):
        """Returns all the plugin topics configured for a provider.

        Plugin topics are defined as configuration keys of a provider's config which indexes a dict with one
        ``'plugin'`` key.

        :param provider: The name of the provider
        :type provider: str or unicode
        :return: The list of supported plugin topics configured for this provider
        :rtype: list(str)
        """
        provider_config = self.providers_config[provider]
        return [topic for topic in provider_config if 'plugin' in provider_config[topic]]

    def _filter_providers(self, topic, product_type_id, selected):
        """Returns a list of providers to be instantiated according to the product type and the topic.

        :param topic: The plugin topic for which the providers should be configured to be selected
        :type topic: str or unicode
        :param product_type_id: The ID of the product type to be supported by the selected providers
        :type product_type_id: str or unicode
        :param selected: A list of providers that the user wants to instantiate
        :type selected: list(tuple(str, int))
        :returns: The providers that support the given product type id and their priorities
        :rtype: list(tuple(str, int))
        """
        if selected is not None:
            return [
                (provider, priority)
                for provider, priority in selected
                if topic in self.providers_config[provider]
            ]
        return [
            # If the priority is not set in the config, the default priority is 0
            (provider, provider_config.get('priority', 0))
            for provider, provider_config in self.providers_config.items()
            if topic in provider_config and product_type_id in provider_config['products']
        ]

    def _instantiate(self, topic, providers=None):
        """Instantiate a set of providers' topic plugins.

        :param topic: The plugin topic for which we should instantiate the providers' plugins (e.g. ``'search'``)
        :type topic: str or unicode
        :param providers: Restrict instantiation to these providers
        :type providers: list(tuple(str, int))
        :return: The instances requested
        :rtype: list
        """
        instances = []
        PluginBaseClass = self._get_base_class(topic)
        if providers is not None:
            selected_providers = [
                (name, self.providers_config[name], priority)
                for name, priority in providers
            ]
        else:
            #  If providers is not given, all providers supporting topic are instantiated with equal priority of 0
            selected_providers = [
                (name, provider_config, 0)
                for name, provider_config in self.providers_config.items()
                if topic in provider_config
            ]
        search_or_api = topic in ('search', 'api')
        if selected_providers:
            logger.debug('Creating configured *%s* plugin instances', topic.upper())
        for provider_name, provider_config, provider_priority in selected_providers:
            logger.debug("Creating '%s' plugin instance with name '%s'", topic.upper(), provider_name)
            # Hack to nest 'products' config key under 'search' or 'api' config key of the provider. This is done
            # to make the mapping between eodag's product type codes and provider's product type codes available to
            # the Search and Api plugins, considering the way plugins are instantiated (they get config[topic] as
            # initialisation parameter. See the instantiate_plugin_by_config method)
            if search_or_api:
                provider_config[topic].setdefault('products', provider_config['products'])
            instance = self.instantiate_plugin_by_config(
                topic,
                provider_config[topic],
                base=PluginBaseClass,
                iname=provider_name,
                priority=provider_priority
            )
            instances.append(instance)
        return instances

    def _get_base_class(self, topic):
        try:
            PluginBaseClass = self.supported_topics[topic]
        except KeyError:
            raise RuntimeError(
                "Topic '{}' is not supported by the plugin system. Choose between {}".format(
                    topic,
                    ', '.join(self.supported_topics.keys())))
        return PluginBaseClass



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import logging
import re

import pyproj
from owslib.csw import CatalogueServiceWeb
from owslib.fes import (
    BBox, PropertyIsEqualTo, PropertyIsGreaterThanOrEqualTo, PropertyIsLessThanOrEqualTo, PropertyIsLike,
)
from owslib.ows import ExceptionReport
from shapely import geometry

from eodag.api.product import EOProduct
from eodag.api.product.representations import properties_from_xml
from eodag.plugins.search.base import Search
from eodag.utils import DEFAULT_PROJ, slugify
from eodag.utils.import_system import patch_owslib_requests


logger = logging.getLogger('beodag.plugins.search.csw')

SUPPORTED_REFERENCE_SCHEMES = [
    'WWW:DOWNLOAD-1.0-http--download',
]

# Configuration keys
SEARCH_DEF = 'search_definitions'
PRODUCT_TYPE = 'pt_tags'
DATE = 'date_tags'


class CSWSearch(Search):

    def __init__(self, config):
        super(CSWSearch, self).__init__(config)
        self.catalog = None

    def query(self, product_type, **kwargs):
        logger.info('New search for product type : %s on %s interface', product_type, self.name)
        auth = kwargs.pop('auth', None)
        if auth is not None:
            self.__init_catalog(**auth.config['credentials'])
        else:
            self.__init_catalog()
        results = []
        if self.catalog:
            for product_type_def in self.config[SEARCH_DEF][PRODUCT_TYPE]:
                product_type_search_tag = product_type_def['name']
                logger.debug('Querying %s for product type %s', product_type_search_tag, product_type)
                constraints = self.__convert_query_params(product_type_def, product_type, kwargs)
                with patch_owslib_requests(verify=False):
                    try:
                        self.catalog.getrecords2(constraints=constraints, esn='full', maxrecords=10)
                    except ExceptionReport:
                        import traceback as tb
                        logger.warning('Failed to query %s for product type %s : %s', product_type_search_tag,
                                       product_type, tb.format_exc())
                        continue
                partial_results = [
                    self.__build_product(record, product_type, **kwargs) for record in self.catalog.records.values()]
                logger.info('Found %s results querying %s', len(partial_results), product_type_search_tag)
                results.extend(partial_results)
        logger.info('Found %s overall results', len(results))
        return results

    def __init_catalog(self, username=None, password=None):
        """Initializes a catalogue by performing a GetCapabilities request on the url"""
        if not self.catalog:
            logger.debug('Initialising CSW catalog at %s', self.config['api_endpoint'])
            with patch_owslib_requests(verify=False):
                try:
                    self.catalog = CatalogueServiceWeb(
                        self.config['api_endpoint'],
                        version=self.config.get('version', '2.0.2'),
                        username=username,
                        password=password
                    )
                except Exception as e:
                    logger.warning('Initialization of catalog failed due to error: (%s: %s)', type(e), e)

    def __build_product(self, rec, product_type, **kwargs):
        """Enable search results to be handled by http download plugin"""
        bbox = rec.bbox_wgs84
        if not bbox:
            code = 'EPSG:4326'
            if rec.bbox.crs.code and rec.bbox.crs.code > 0:
                code = ':'.join((str(rec.bbox.crs.id), str(rec.bbox.crs.code)))
            rec_proj = pyproj.Proj(init=code)
            maxx, maxy = pyproj.transform(rec_proj, DEFAULT_PROJ, rec.bbox.maxx, rec.bbox.maxy)
            minx, miny = pyproj.transform(rec_proj, DEFAULT_PROJ, rec.bbox.minx, rec.bbox.miny)
            bbox = (minx, miny, maxx, maxy)
        download_url = ''
        resource_filter = re.compile(self.config[SEARCH_DEF].get('resource_location_filter', ''))
        for ref in rec.references:
            if ref['scheme'] in SUPPORTED_REFERENCE_SCHEMES:
                if resource_filter.pattern and resource_filter.search(ref['url']):
                    download_url = ref['url']
                else:
                    download_url = ref['url']
                break
        properties = properties_from_xml(rec.xml, self.config['metadata_mapping'])
        properties['title'] = slugify(rec.identifier)
        properties['geometry'] = geometry.box(*bbox)
        properties['productType'] = product_type
        return EOProduct(
            product_type,
            self.instance_name,
            download_url,
            properties,
            searched_bbox=kwargs.get('footprints'),
        )

    def __convert_query_params(self, product_type_def, product_type, params):
        """Translates eodag search to CSW constraints using owslib constraint classes"""
        constraints = []
        # How the match should be performed (fuzzy, prefix, postfix or exact). defaults to fuzzy
        pt_tag, matching = product_type_def['name'], product_type_def.get('matching', 'fuzzy')
        if matching == 'prefix':
            constraints.append(PropertyIsLike(pt_tag, '{}%'.format(product_type)))
        elif matching == 'postfix':
            constraints.append(PropertyIsLike(pt_tag, '%{}'.format(product_type)))
        elif matching == 'exact':
            constraints.append(PropertyIsEqualTo(pt_tag, product_type))
        else:  # unknown matching is considered to be equal to 'fuzzy'
            constraints.append(PropertyIsLike(pt_tag, '%{}%'.format(product_type)))

        # footprint
        fp = params.get('geometry')
        if fp:
            constraints.append(BBox([fp['lonmin'], fp['latmin'], fp['lonmax'], fp['latmax']]))

        # dates
        start, end = params.get('startTimeFromAscendingNode'), params.get('completionTimeFromAscendingNode')
        if start:
            constraints.append(PropertyIsGreaterThanOrEqualTo(self.config[SEARCH_DEF][DATE]['start'], start))
        if end:
            constraints.append(PropertyIsLessThanOrEqualTo(self.config[SEARCH_DEF][DATE]['end'], end))
        return [constraints] if len(constraints) > 1 else constraints



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

import logging

from shapely import geometry

try:
    from shapely.errors import TopologicalError
except ImportError:
    from shapely.geos import TopologicalError

from eodag.plugins.crunch.base import Crunch


logger = logging.getLogger('eodag.plugins.crunch.filter_overlap')


class FilterOverlap(Crunch):

    def proceed(self, products, **search_params):
        """Filter products, retaining only those that are overlapping with the search_extent"""
        logger.info('Start filtering for overlapping products')
        filtered = []
        add_to_filtered = filtered.append
        footprint = search_params.get('geometry')
        if not footprint:
            return products
        minimum_overlap = float(self.config.get('minimum_overlap', '0'))
        search_extent = geometry.box(footprint['lonmin'], footprint['latmin'],
                                     footprint['lonmax'], footprint['latmax'])
        logger.info('Initial requested extent area: %s', search_extent.area)
        for product in products:
            logger.info('Uncovered extent area: %s', search_extent.area)
            if product.search_intersection:
                intersection = product.search_intersection
                product_geometry = product.geometry
            else:  # Product geometry may be invalid
                if not product.geometry.is_valid():
                    logger.debug('Trying our best to deal with invalid geometry on product: %r', product)
                    product_geometry = product.geometry.buffer(0)
                    try:
                        intersection = search_extent.intersection(product_geometry)
                    except TopologicalError:
                        logger.debug('Product geometry still invalid. Overlap test restricted to containment')
                        if search_extent.contains(product_geometry):
                            logger.debug('Product %r overlaps the search extent. Adding it to filtered results')
                            add_to_filtered(product)
                        continue
                else:
                    product_geometry = product.geometry
                    intersection = search_extent.intersection(product_geometry)
            ipos = (intersection.area / search_extent.area) * 100
            ipop = (intersection.area / product_geometry.area) * 100
            logger.info('Intersection of product extent and search extent covers %f percent of the search extent '
                        'area', ipos)
            logger.info('Intersection of product extent and search extent covers %f percent of the product extent '
                        'area', ipop)
            if any((search_extent.contains(product.geometry), ipos >= minimum_overlap, ipop >= minimum_overlap)):
                logger.info('Product %r overlaps the search extent. Adding it to filtered results', product)
                add_to_filtered(product)
        logger.info('Finished filtering products. Resulting products: %r', filtered)
        return filtered



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

import jsonpath_rw as jsonpath
from lxml import etree

from eodag.utils.metadata_mapping import get_metadata_path


DEFAULT_METADATA_MAPPING = {
    'id': '$.id',
    'geometry': '$.geometry',
    'productType': '$.properties.productType',
    'platform': '$.properties.platform',
    'platformSerialIdentifier': '$.properties.platformSerialIdentifier',
    'instrument': '$.properties.instrument',
    'sensorType': '$.properties.sensorType',
    'processingLevel': '$.properties.processingLevel',
    'orbitType': '$.properties.orbitType',
    'title': '$.properties.title',
    'topicCategory': '$.properties.topicCategory',
    'keyword': '$.properties.keyword',
    'abstract': '$.properties.abstract',
    'organisationName': '$.properties.organisationName',
    'orbitNumber': '$.properties.orbitNumber',
    'orbitDirection': '$.properties.orbitDirection',
    'cloudCover': '$.properties.cloudCover',
    'snowCover': '$.properties.snowCover',
    'startTimeFromAscendingNode': '$.properties.startTimeFromAscendingNode',
    'completionTimeFromAscendingNode': '$.properties.completionTimeFromAscendingNode',
    'parentIdentifier': '$.properties.parentIdentifier',
    'resolution': '$.properties.resolution',
    'sensorMode': '$.properties.sensorMode',
    'quicklook': '$.properties.quicklook',
}


def properties_from_json(json, mapping):
    """Extract properties from a provider json result.

    :param json: the representation of a provider result as a json object
    :type json: dict
    :param mapping: a mapping between :class:`~eodag.api.product.EOProduct`'s metadata keys and the location of the
                    values of these properties in the json representation, expressed as a
                    `jsonpath <http://goessner.net/articles/JsonPath/>`_
    :return: the metadata of the :class:`~eodag.api.product.EOProduct`
    :rtype: dict
    """
    properties = {}
    for metadata in DEFAULT_METADATA_MAPPING:
        if metadata not in mapping:
            properties[metadata] = 'N/A'
        else:
            path = jsonpath.parse(get_metadata_path(mapping[metadata]))
            match = path.find(json)
            properties[metadata] = match[0].value if len(match) == 1 else None
    return properties


def properties_from_xml(xml_as_text, mapping):
    """Extract properties from a provider xml result.

    :param xml_as_text: the representation of a provider result as xml
    :type xml_as_text: str or unicode
    :param mapping: a mapping between :class:`~eodag.api.product.EOProduct`'s metadata keys and the location of the
                    values of these properties in the xml representation, expressed as a
                    `xpath <https://www.w3schools.com/xml/xml_xpath.asp>`_
    :return: the metadata of the :class:`~eodag.api.product.EOProduct`
    :rtype: dict
    """
    properties = {}
    root = etree.XML(xml_as_text)
    for metadata in DEFAULT_METADATA_MAPPING:
        if metadata not in mapping:
            properties[metadata] = 'N/A'
        else:
            value = root.xpath(get_metadata_path(mapping[metadata]), namespaces=root.nsmap)
            if len(value) > 1:
                properties[metadata] = value
            else:
                properties[metadata] = value[0] if len(value) == 1 else None
    return properties



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

import string
from random import SystemRandom

import requests
from lxml import etree
from requests.auth import AuthBase

from eodag.plugins.authentication import Authentication
from eodag.utils import parse_qs, repeatfunc, urlencode, urlparse, urlunparse
from eodag.utils.exceptions import AuthenticationError, MisconfiguredError


class OIDCAuthorizationCodeFlowAuth(Authentication):
    """Implement the authorization code flow of the OpenIDConnect authorization specification.

    The `OpenID Connect <http://openid.net/specs/openid-connect-core-1_0.html>`_ specification
    adds an authentication layer on top of oauth 2.0. This plugin implements the
    `authorization code flow <http://openid.net/specs/openid-connect-core-1_0.html#Authentication>`_
    option of this specification.
    The particularity of this plugin is that it proceeds to a headless (not involving the user)
    interaction with the `Theia <https://sso.theia-land.fr/>`_ OpenID provider to authenticate a
    registered user with its username and password on the server and then granting to eodag the
    necessary rights. It does that using the client ID of the eodag provider that use it.

    The configuration keys of this plugin are as follows (they have no defaults)::

        # The authorization url of the server (where to query for grants)
        authorization_uri:

        # The callback url that will handle the code given by the OIDC provider
        redirect_uri:

        # The url of the authentication backend of the OIDC provider
        authentication_uri:

        # The url to query to exchange the authorization code obtained from the OIDC provider
        # for an authorized token
        token_uri:

        # The OIDC provider's client ID of the eodag provider
        client_id:

        # An optional mapping between OIDC url query string and token handler query string
        # params (only necessary if they are not the same as for OIDC). This is eodag provider
        # dependant
        token_exchange_params:
          redirect_uri: redirectUri
          client_id: clientId

        # One of: json, data or params. This is the way to pass the data to the POST request
        # that is made to the token server. They correspond to the recognised keywords arguments
        # of the Python `requests <http://docs.python-requests.org/>`_ library
        token_exchange_post_data_method:

        # One of qs or header. This is how the token obtained will be used to authenticate the user
        # on protected requests. if 'qs' is chosen, then 'token_qs_key' is mandatory
        token_provision:

        # Only necessary when 'token_provision' is 'qs'. Refers to the name of the query param to be
        # used in the query request
        token_qs_key:

    """
    SCOPE = 'openid'
    RESPONSE_TYPE = 'code'

    def __init__(self, config):
        super(OIDCAuthorizationCodeFlowAuth, self).__init__(config)
        self.session = requests.Session()

    def authenticate(self):
        state = self.compute_state()
        params = {
            'client_id': self.config['client_id'],
            'response_type': self.RESPONSE_TYPE,
            'scope': self.SCOPE,
            'state': state,
            'redirect_uri': self.config['redirect_uri'],
        }
        authentication_response = self.authenticate_user(self.session.get(
            self.config['authorization_uri'],
            params=params
        ))
        user_consent_response = self.grant_user_consent(authentication_response)
        try:
            token = self.exchange_code_for_token(user_consent_response.url, state)
        except Exception:
            import traceback as tb
            raise AuthenticationError(
                'Something went wrong while trying to get authorization token:\n{}'.format(tb.format_exc())
            )
        if self.config['token_provision'] not in ('qs', 'headers'):
            raise MisconfiguredError('Provider config parameter "token_provision" must be one of "qs" or "headers"')
        if self.config['token_provision'] == 'qs' and not self.config.get('token_qs_key', ''):
            raise MisconfiguredError('Provider config parameter "token_provision" with value "qs" must have '
                                     '"token_qs_key" config parameter as well')
        return CodeAuthorizedAuth(token, self.config['token_provision'], key=self.config.get('token_qs_key'))

    def authenticate_user(self, authorization_response):
        login_document = etree.HTML(authorization_response.text)
        login_form = login_document.xpath('//form[@id="loginForm"]')[0]
        login_data = self.config['credentials']
        login_data['sessionDataKey'] = login_form.xpath('//input[@name="sessionDataKey"]')[0].attrib['value']
        return self.session.post(self.config['authentication_uri'], data=login_data)

    def grant_user_consent(self, authentication_response):
        user_consent_document = etree.HTML(authentication_response.text)
        user_consent_form = user_consent_document.xpath('//form[@id="profile"]')[0]
        user_consent_data = {
            'consent': 'approve',
            'sessionDataKeyConsent': user_consent_form.xpath(
                '//input[@name="sessionDataKeyConsent"]'
            )[0].attrib['value']
        }
        return self.session.post(self.config['authorization_uri'], data=user_consent_data)

    def exchange_code_for_token(self, authorized_url, state):
        qs = parse_qs(urlparse(authorized_url).query)
        if qs['state'][0] != state:
            raise AuthenticationError(
                'The state received in the authorized url does not match initially computed state')
        code = qs['code'][0]
        token_exchange_data = {
            'redirect_uri': self.config['redirect_uri'],
            'client_id': self.config['client_id'],
            'code': code,
            'state': state,
        }
        custom_token_exchange_params = self.config['token_exchange_params']
        if custom_token_exchange_params:
            token_exchange_data[custom_token_exchange_params['redirect_uri']] = token_exchange_data.pop('redirect_uri')
            token_exchange_data[custom_token_exchange_params['client_id']] = token_exchange_data.pop('client_id')
        post_request_kwargs = {self.config['token_exchange_post_data_method']: token_exchange_data}
        r = self.session.post(self.config['token_uri'], **post_request_kwargs)
        return r.json()['token']

    @staticmethod
    def compute_state():
        rand = SystemRandom()
        return ''.join(repeatfunc(rand.choice, 22, string.digits + string.ascii_lowercase + string.ascii_uppercase))


class CodeAuthorizedAuth(AuthBase):

    def __init__(self, token, where, key=None):
        self.token = token
        self.where = where
        self.key = key

    def __call__(self, request):
        if self.where == 'qs':
            parts = urlparse(request.url)
            qs = parse_qs(parts.query)
            qs[self.key] = self.token
            request.url = urlunparse((
                parts.scheme,
                parts.netloc,
                parts.path,
                parts.params,
                urlencode(qs),
                parts.fragment
            ))
        elif self.where == 'header':
            request.headers['Authorization'] = "Bearer {}".format(self.token)
        return request



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

import logging

import requests

from eodag.api.product import EOProduct
from eodag.api.product.representations import properties_from_json
from eodag.plugins.search.base import Search
from eodag.utils import get_timestamp
from eodag.utils.metadata_mapping import get_search_param


logger = logging.getLogger('eodag.plugins.search.arlas')


class ArlasSearch(Search):
    SEARCH_ENDPOINT_TPL = '/explore/{arlas_collection}/_geosearch'
    COUNT_ENDPOINT_TPL = '/explore/{arlas_collection}/_count'

    def query(self, product_type, auth=None, **kwargs):
        logger.info('New search for product type : *%s* on %s interface', product_type, self.name)
        query_string = self.build_query_string(product_type, kwargs)
        if auth:
            auth = auth.authenticate()

        total_hits = self.count_products(query_string, auth)

        # Do not perform the search if the criteria lead to 0 result
        if total_hits == 0:
            return []

        # If there was a silenced error during result counting (total_hits == -1), perform the search anyway without
        # specifying the size query string parameter
        if total_hits != -1:
            query_string += '&size={}'.format(total_hits)

        search_url = '{base}{path}?{qs}'.format(
            base=self.config['api_endpoint'].rstrip('/'),
            path=self.SEARCH_ENDPOINT_TPL.format(**self.config),
            qs=query_string
        )

        logger.info('Making search request at: %s', search_url)
        response = requests.get(search_url, auth=auth)
        try:
            logger.info('Checking response')
            response.raise_for_status()
            logger.info('Search requests successful. HTTP status code: %s', response.status_code)
        except requests.HTTPError:
            import traceback as tb
            logger.warning('Failed to query arlas server at %s. Got error:\n%s',
                           self.config['api_endpoint'], tb.format_exc())
            return []

        results = response.json()
        products = []
        try:
            if results['features']:
                logger.info('Normalizing results')
                for feature in results['features']:
                    properties = properties_from_json(feature, self.config['metadata_mapping'])
                    # In a provider config using this plugin, quicklook maps to a property which
                    # gives information whether the product has a quicklook or not. Based on this
                    # info, we can build the quicklook url
                    properties['quicklook'] = (
                        '{}/{}'.format(self.config['quicklook_endpoint'], properties['id'])
                        if properties['quicklook']
                        else
                        None
                    )
                    products.append(EOProduct(
                        product_type,
                        self.instance_name,
                        '{base}' + '/{}'.format(properties['id']),
                        properties,
                        searched_bbox=kwargs.get('footprint')
                    ))
        except KeyError as ke:
            if 'features' in ke:
                logger.warning('Invalid geojson returned: %s, assuming no results', results)
                return []
            raise ke
        logger.info('Search on Arlas server %s succeeded. Results: %s', self.config['api_endpoint'], products)
        return products

    def build_query_string(self, product_type, options):
        """Build the query string that is used to perform search requests

        :param product_type: the code of the product type to look for as defined by eodag
        :type product_type: str or unicode
        :param options: the additional search criteria that are not mandatory for a search on eodag
        :type options: dict
        :return:
        """
        product_type_param = get_search_param(self.config['metadata_mapping']['productType'])
        cloud_cover_param = get_search_param(self.config['metadata_mapping']['cloudCover'])
        geometry_param = get_search_param(self.config['metadata_mapping']['geometry'])
        start_date_param = get_search_param(self.config['metadata_mapping']['startTimeFromAscendingNode'])

        logger.debug('Building the query string that will be used for search')
        mandatory_qs = 'f={}:eq:{}'.format(product_type_param, self.config['products'][product_type]['product_type'])
        optional_qs = ''

        max_cloud_cover = options.get('cloudCover')
        if max_cloud_cover:
            logger.debug('Adding filter for max cloud cover: %s', max_cloud_cover)
            optional_qs += '&{}:range:0,{max}'.format(cloud_cover_param, max=max_cloud_cover)

        footprint = options.get('geometry')
        if footprint:
            logger.debug('Adding filter for footprint: %s', footprint)
            optional_qs += '&{}={lonmin},{latmin},{lonmax},{latmax}'.format(geometry_param, **footprint)

        start_date = options.get('startTimeFromAscendingNode')
        end_date = options.get('completionTimeFromAscendingNode')
        if start_date:
            start_timestamp = int(1e3 * get_timestamp(start_date))
            if end_date:
                logger.debug('Adding filter for sensing date range: %s - %s', start_date, end_date)
                end_timestamp = int(1e3 * get_timestamp(end_date))
                optional_qs += '&f={}:range:[{min}<{max}]'.format(
                    start_date_param, min=start_timestamp, max=end_timestamp)
            else:
                logger.debug('Adding filter for minimum sensing date: %s', start_date)
                optional_qs += '&f={}:gte:{min}'.format(start_date_param, min=start_timestamp)
        elif end_date:
            logger.debug('Adding filter for maximum sensing date: %s', end_date)
            end_timestamp = int(1e3 * get_timestamp(end_date))
            optional_qs += '&f={}:lte:{max}'.format(start_date_param, max=end_timestamp)

        return '{}{}'.format(mandatory_qs, optional_qs)

    def count_products(self, query_string, auth):
        """Return the total number of results that can be found given the search_url

        :param query_string: the query string that will serve as filter
        :type query_string: str or unicode
        :param auth: the authentication information if needed
        :return: the number of products that will be reached if we perform a search with this query string
        :rtype: int
        """
        logger.info('Looking for the number of products satisfying the search criteria')
        response = requests.get(
            '{base}{path}?{qs}'.format(
                base=self.config['api_endpoint'],
                path=self.COUNT_ENDPOINT_TPL.format(**self.config),
                qs=query_string
            ),
            auth=auth
        )
        try:
            response.raise_for_status()
        except requests.HTTPError:
            import traceback as tb
            logger.warning('Unable to determine the number of products satisfying the criteria. Got error:\n%s',
                           tb.format_exc())
            return -1
        return response.json()['totalnb']



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import hashlib
import logging
import os
import zipfile

import requests
from requests import HTTPError
from tqdm import tqdm

from eodag.plugins.download.base import Download
from eodag.plugins.search.base import MisconfiguredError


logger = logging.getLogger('eodag.plugins.download.http')


class HTTPDownload(Download):

    def __init__(self, config):
        super(HTTPDownload, self).__init__(config)
        if 'base_uri' not in self.config:
            raise MisconfiguredError('{} plugin require a base_uri configuration key'.format(self.name))
        self.config.setdefault('outputs_prefix', '/tmp')
        self.config.setdefault('extract', True)
        logger.debug('Images will be downloaded to directory %s', self.config['outputs_prefix'])

    def download(self, product, auth=None, progress_callback=None):
        """Download a product as zip archive using HTTP protocol"""
        if not product.location.startswith('file://'):
            url = self.__build_download_url(product, auth)
            if not url:
                logger.debug('Unable to get download url for %s, skipping download', product)
                return
            logger.debug('Download url: %s', url)

            # Strong asumption made here: all products downloaded will be zip archives
            filename = product.properties['title'] + '.zip'
            local_file_path = os.path.join(self.config['outputs_prefix'], filename)
            download_records = os.path.join(self.config['outputs_prefix'], '.downloaded')
            try:
                os.makedirs(download_records)
            except OSError as exc:
                import errno
                if exc.errno != errno.EEXIST:   # Skip error if dir exists
                    import traceback as tb
                    logger.warning('Unable to create record directory. Got:\n%s', tb.format_exc())
            url_hash = hashlib.md5(url.encode('utf-8')).hexdigest()
            record_filename = os.path.join(download_records, url_hash)
            if os.path.isfile(record_filename) and os.path.isfile(local_file_path):
                return self.__finalize(local_file_path)
            # Remove the record file if local_file_path is absent (e.g. it was deleted while record wasn't)
            elif os.path.isfile(record_filename):
                logger.debug('Record file found (%s) but not the actual file', record_filename)
                logger.debug('Removing record file : %s', record_filename)
                os.remove(record_filename)

            with requests.get(url, stream=True, auth=auth, params=self.config.get('dl_url_params', {})) as stream:
                try:
                    stream.raise_for_status()
                except HTTPError:
                    import traceback as tb
                    logger.error("Error while getting resource :\n%s", tb.format_exc())
                else:
                    stream_size = int(stream.headers.get('content-length', 0))
                    with open(local_file_path, 'wb') as fhandle:
                        for chunk in stream.iter_content(chunk_size=64 * 1024):
                            if chunk:
                                fhandle.write(chunk)
                                progress_callback(len(chunk), stream_size)

                    with open(record_filename, 'w') as fh:
                        fh.write(url)
                    logger.debug('Download recorded in %s', record_filename)
                    return self.__finalize(local_file_path)
        else:
            path = product.location.replace('file://', '')
            logger.info('Product already present on this platform. Identifier: %s', path)
            # Do not download data if we are on site. Instead give back the absolute path to the data
            return path

    def __build_download_url(self, product, auth):
        if product.location:
            try:
                url = product.location.format(
                    base=self.config.get('base_uri').rstrip('/'),
                )
                if product.properties['organisationName'] in ('ESA',):
                    url += '?token={}'.format(auth.token)
                return url
            except Exception:
                import traceback as tb
                raise RuntimeError('Product {} is incompatible with download plugin {}\n{}'.format(
                    product, self.name, tb.format_exc()
                ))

    def __finalize(self, local_file_path):
        """Finalize the download process.

        :param local_file_path: The path to the local zip archive downloaded or already present
        :type local_file_path: str or unicode
        :return: the absolute path to the product
        """
        if self.config['extract'] and zipfile.is_zipfile(local_file_path):
            if not os.path.exists(local_file_path[:local_file_path.index('.zip')]):
                logger.info('Extraction activated')
                with zipfile.ZipFile(local_file_path, 'r') as zfile:
                    fileinfos = zfile.infolist()
                    with tqdm(fileinfos, unit='file', desc='Extracting files from {}'.format(
                            local_file_path)) as progressbar:
                        for fileinfo in progressbar:
                            zfile.extract(
                                fileinfo,
                                path=os.path.join(
                                    self.config['outputs_prefix'],
                                    local_file_path[:local_file_path.index('.zip')])
                            )
            # Handle depth levels in the product archive. For example, if the downloaded archive was
            # extracted to: /top_level/product_base_dir and archive_depth was configured to 2, the product
            # location will be /top_level/product_base_dir.
            # WARNING: A strong assumption is made here: there is only one subdirectory per level
            archive_depth = self.config.get('archive_depth', 1)
            product_path = local_file_path[:local_file_path.index('.zip')]
            count = 1
            while count < archive_depth:
                product_path = os.path.join(product_path, os.listdir(product_path)[0])
                count += 1
            return product_path
        else:
            return local_file_path



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from setuptools import setup, find_packages

BASEDIR = os.path.dirname(os.path.abspath(os.path.realpath(__file__)))

metadata = {}
with open(os.path.join(BASEDIR, 'eodag', '__meta__.py'), 'r') as f:
    exec(f.read(), metadata)

with open(os.path.join(BASEDIR, 'README.rst'), 'r') as f:
    readme = f.read()

setup(
    name=metadata['__title__'],
    version=metadata['__version__'],
    description=metadata['__description__'],
    long_description=readme,
    author=metadata['__author__'],
    author_email=metadata['__author_email__'],
    url=metadata['__url__'],
    license=metadata['__license__'],
    packages=find_packages(exclude=('*.tests', '*.tests.*', 'tests.*', 'tests')),
    package_data={'': ['LICENSE', 'NOTICE']},
    include_package_data=True,
    install_requires=[
        'click',
        'requests',
        'python-dateutil',
        'PyYAML',
        'tqdm',
        'shapely',
        'owslib',
        'six',
        'geojson',
        'pyproj',
        'usgs',
        'boto3',
        'numpy',
        'rasterio==1.0a12',
        'protobuf',
        'grpcio',
        # To be able to do 'import concurrent.futures' in Python 2.7
        "futures; python_version < '3.5'",
        'jsonpath-rw',
        'lxml',
        'xarray',
    ],
    extras_require={
        'dev': [
            'nose',
            'tox',
            'faker',
            'mock; python_version < "3.5" ',
            'coverage',
            'moto',
            'twine',
            'wheel',
            'flake8',
        ],
        'tutorials': [
            'jupyter',
            'ipyleaflet',
            'ipywidgets',
            'matplotlib',
        ],
        'docs': [
            'sphinx',
            'nbsphinx',
        ]
    },
    entry_points={
        'console_scripts': [
            'eodag = eodag.cli:eodag'
        ],
        'eodag.plugins.api': [
            'UsgsApi = eodag.plugins.apis.usgs:UsgsApi',
        ],
        'eodag.plugins.auth': [
            'GenericAuth = eodag.plugins.authentication.generic:GenericAuth',
            'HTTPHeaderAuth = eodag.plugins.authentication.header:HTTPHeaderAuth',
            'OAuth = eodag.plugins.authentication.oauth:OAuth',
            'TokenAuth = eodag.plugins.authentication.token:TokenAuth',
            'OIDCAuthorizationCodeFlowAuth = eodag.plugins.authentication.openid_connect:OIDCAuthorizationCodeFlowAuth',    # noqa
        ],
        'eodag.plugins.crunch': [
            'FilterLatestIntersect = eodag.plugins.crunch.filter_latest_intersect:FilterLatestIntersect',
            'FilterLatestByName = eodag.plugins.crunch.filter_latest_tpl_name:FilterLatestByName',
            'FilterOverlap = eodag.plugins.crunch.filter_overlap:FilterOverlap',
        ],
        'eodag.plugins.download': [
            'AwsDownload = eodag.plugins.download.aws:AwsDownload',
            'HTTPDownload = eodag.plugins.download.http:HTTPDownload',
        ],
        'eodag.plugins.search': [
            'ArlasSearch = eodag.plugins.search.arlas:ArlasSearch',
            'AwsSearch = eodag.plugins.search.aws:AwsSearch',
            'CSWSearch = eodag.plugins.search.csw:CSWSearch',
            'RestoSearch = eodag.plugins.search.resto:RestoSearch',
        ],
    },
    project_urls={
        "Bug Tracker": "https://bitbucket.org/geostorm/eodag/issues/",
        "Documentation": "https://eodag.readthedocs.io/en/latest/",
        "Source Code": "https://bitbucket.org/geostorm/eodag",
    },
    classifiers=(
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering :: GIS',
    ),
)



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import logging
import re

import requests
from requests import HTTPError

from eodag.api.product import EOProduct
from eodag.api.product.representations import properties_from_json
from eodag.utils import urljoin, urlparse
from eodag.utils.metadata_mapping import get_search_param
from .base import Search


logger = logging.getLogger('eodag.plugins.search.resto')


class RestoSearch(Search):
    SEARCH_PATH = '/collections/{collection}/search.json'

    def __init__(self, config):
        super(RestoSearch, self).__init__(config)
        self.query_url_tpl = urljoin(
            self.config['api_endpoint'],
            urlparse(self.config['api_endpoint']).path.rstrip('/') + self.SEARCH_PATH
        )
        # What scheme is used to locate the products that will be discovered during search
        self.product_location_scheme = self.config.get('product_location_scheme', 'https')

    def query(self, product_type, auth=None, **kwargs):
        logger.info('New search for product type : *%s* on %s interface', product_type, self.name)
        results = []
        add_to_results = results.extend
        start_date_param = get_search_param(self.config['metadata_mapping']['startTimeFromAscendingNode'])
        end_date_param = get_search_param(self.config['metadata_mapping']['completionTimeFromAscendingNode'])
        product_type_param = get_search_param(self.config['metadata_mapping']['productType'])
        cloud_cover_param = get_search_param(self.config['metadata_mapping']['cloudCover'])
        geometry_param = get_search_param(self.config['metadata_mapping']['geometry'])
        params = {
            'sortOrder': 'descending',
            'sortParam': start_date_param,
            start_date_param: kwargs.pop('startTimeFromAscendingNode', None),
        }
        cloud_cover = kwargs.pop('cloudCover', None)
        if cloud_cover is not None:
            if not 0 <= cloud_cover <= 100:
                raise RuntimeError("Invalid cloud cover criterium: '{}'. Should be a percentage (bounded in [0-100])")
            params[cloud_cover_param] = '[0,{}]'.format(cloud_cover)
        end_date = kwargs.pop('completionTimeFromAscendingNode', None)
        if end_date:
            params[end_date_param] = end_date
        footprint = kwargs.pop('geometry', None)
        if footprint:
            params[geometry_param] = '{lonmin},{latmin},{lonmax},{latmax}'.format(**footprint)

        collections, resto_product_type = self.map_product_type(product_type, params[start_date_param])
        # len(collections) == 2 If and Only if the product type is S2-L1C, provider is PEPS and there is no search
        # constraint on date. Otherwise, it's equal to 1
        for collection in collections:
            logger.debug('Collection found for product type %s: %s', product_type, collection)
            logger.debug('Corresponding Resto product_type found for product type %s: %s',
                         product_type, resto_product_type)
            params[product_type_param] = resto_product_type

            url = self.query_url_tpl.format(collection=collection)
            logger.debug('Making request to %s with params : %s', url, params)
            try:
                response = requests.get(url, params=params)
                response.raise_for_status()
            except HTTPError as e:
                logger.debug('Skipping error while searching for %s RestoSearch instance product type %s: %s',
                             self.instance_name, resto_product_type, e)
            else:
                add_to_results(self.normalize_results(product_type, response.json(), footprint))
        return results

    def map_product_type(self, product_type, date):
        """Maps the eodag's specific product type code to Resto specific product type (which is a collection id and a
        product type id)

        :param product_type: The eodag specific product type code name
        :type product_type: str or unicode
        :param date: The date search constraint (used only for peps provider)
        :type date: str or unicode
        :return: The corresponding collection and product type ids on this instance of Resto
        :rtype: tuple(tuple, str)
        """
        mapping = self.config['products'][product_type]
        # See https://earth.esa.int/web/sentinel/missions/sentinel-2/news/-/asset_publisher/Ac0d/content/change-of
        # -format-for-new-sentinel-2-level-1c-products-starting-on-6-december
        if product_type == 'S2_MSI_L1C':
            if self.instance_name == 'peps':
                # If there is no criteria on date, we want to query all the collections known for providing L1C
                # products
                if date is None:
                    collection = ('S2', 'S2ST')
                else:
                    match = re.match(r'(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})', date).groupdict()
                    year, month, day = int(match['year']), int(match['month']), int(match['day'])
                    if year == 2016 and month <= 12 and day <= 5:
                        collection = ('S2',)
                    else:
                        collection = ('S2ST',)
            else:
                collection = (mapping['collection'],)
        else:
            collection = (mapping['collection'],)
        return collection, mapping['product_type']

    def normalize_results(self, product_type, results, search_bbox):
        normalized = []
        if results['features']:
            logger.info('Found %s products', len(results['features']))
            logger.debug('Adapting plugin results to eodag product representation')
            for result in results['features']:
                if self.product_location_scheme == 'file':
                    download_url = '{}://{}'.format(
                        self.product_location_scheme,
                        result['properties']['productIdentifier'])
                else:
                    if result['properties']['organisationName'] in ('ESA',):
                        # TODO: See the above todo about that productIdentifier thing
                        download_url = '{base}' + '/{prodId}.zip'.format(
                            prodId=result['properties']['productIdentifier'].replace('/eodata/', '')
                        )
                    else:
                        if result['properties'].get('services', {}).get('download', {}).get('url'):
                            download_url = result['properties']['services']['download']['url']
                        else:
                            download_url = '{base}' + '/collections/{collection}/{feature_id}/download'.format(
                                collection=result['properties']['collection'],
                                feature_id=result['id'],
                            )
                product = EOProduct(
                    product_type,
                    self.instance_name,
                    download_url,
                    properties_from_json(result, self.config['metadata_mapping']),
                    searched_bbox=search_bbox
                )
                normalized.append(product)
            logger.debug('Normalized products : %s', normalized)
        else:
            logger.info('Nothing found !')
        return normalized
